/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: propgroup.c,v 1.4 1998/06/05 19:21:50 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <proplist.h>
#include <gtk/gtk.h>
#include "wtype.h"
#include "propgroup.h"
#include "proputil.h"


PropertyGroup *PropertyGroupCreate( void )
{
    PropertyGroup *group = malloc( sizeof( PropertyGroup ));

    if( group )
    {
	memset( group, 0, sizeof( PropertyGroup ));
    }
    return group;
}

void PropertyGroupDestroy( PropertyGroup *group )
{

    if( group )
    {
	if( group->entries )
	    free( group->entries );
	if( group->name )
	    PLRelease( group->name );
	if( group->gprops )
	    PLRelease( group->gprops );
	free( group );
    }
}


static int prop_compare( const void *v1, const void *v2 )
{
    const PropertyEntry *p1 = v1;
    const PropertyEntry *p2 = v2;

    return PLOrderCompare( p1->meta, p2->meta );
}


void PropertyGroupSort( PropertyGroup *group )
{

    if( group )
    {
	qsort( group->entries, group->numProps, sizeof( PropertyEntry ),
	       prop_compare );
    }
}


proplist_t PropertyGroupDefaultProps( proplist_t name, int order )
{
    proplist_t	props;
    char	buffer[ 8 ];
    
    props = PLMakeDictionaryFromEntries( NULL, NULL );
    if( props )
    {
	PLDictAdd1( props, "Description", name );
	PLDictAdd2( props, "Help", "None" );
	sprintf( buffer, "%d", order );
	PLDictAdd2( props, "Order", buffer );
    }
    return props;
}

PropertyEntry *PropertyGroupNewEntry( PropertyGroup *group )
{
    PropertyEntry	*ent;
    int			size;

    if( group->numProps >= group->numPropsAllocated ) {
	size = group->numPropsAllocated + 16;
	if( group->entries ) 
	    ent = realloc( group->entries, size * sizeof( PropertyEntry ));
	else
	    ent = malloc( size * sizeof( PropertyEntry ));
	if( ent ) {
	    group->entries = ent;
	    group->numPropsAllocated = size;
	}
	else {
	    return NULL;
	}
    }
    return( &group->entries[group->numProps++] );
}

PropertyEntry *PropertyGroupFindEntry( PropertyGroup *group, proplist_t key )
{
    PropertyEntry	*ent;
    int			i;

    for( i = 0; i < group->numProps; i++ )
    {
	ent = &group->entries[i];

	if( PLIsEqual( ent->key, key ))
	    return ent;
    }
    return NULL;
}


static void table_attach( GtkWidget *table, GtkWidget *widget,
			  gint col, gint row )
{
    gtk_table_attach ( GTK_TABLE( table ), widget,
		       col, col + 1,
		       row, row + 1,
		       GTK_EXPAND | GTK_FILL,
		       GTK_FILL,
		       0, 0);
}
    
char *PropertyGroupGetGprop( PropertyGroup *group, char *name )
{
    proplist_t val;

    /* XXX */
    val = PLGetDictionaryEntry( group->gprops, PLMakeString( name ));
    return PLGetString( val );
}

char *PropertyEntryLabel( PropertyEntry *ent )
{
    proplist_t	prop = NULL;

    /* Try for name from meta properties first */
    if( ent->meta )
	prop = PLGetDictionaryEntry( ent->meta, PLMakeString( "Name" ));
    /* Else default to property name */
    if( !prop )
	prop = ent->key;
    if( prop )
	return PLGetString( prop );
    return NULL;
}

char *PropertyEntryDescription( PropertyEntry *ent )
{
    proplist_t	prop = NULL;
    char	*str;

    if( ent->meta )
    {
	prop = PLGetDictionaryEntry( ent->meta, PLMakeString( "Description" ));
	if( prop )
	{
	    str = PLGetString( prop );
	    /* Don't return anything if it's not set */
	    if( str && strcmp( str, "None" )) 
		return str;
	}
    }
    return NULL;
}


GtkWidget *PropertyGroupWidgetsBuild( PropertyGroup *group )
{
    GtkWidget	*widget;
    GtkWidget	*table;
    GtkTooltips	*tooltips;
    int		row, i;
    char	*labname, *desc;
    PropertyEntry *entry;

    table = gtk_table_new( 10, 2, FALSE );
    gtk_table_set_row_spacings( GTK_TABLE(table), 4 );
    gtk_table_set_col_spacings( GTK_TABLE(table), 2 );
    gtk_widget_show( table );

    tooltips = gtk_tooltips_new();
    
    row = 0;
    for( i = 0; i < group->numProps ; i++ ) {
	entry = &group->entries[i];
	
	entry->widget = NULL;
	if( entry->type == NULL )
	    continue;

	labname = PropertyEntryLabel( entry );
	desc = PropertyEntryDescription( entry );
	if( labname ) {
	    widget = gtk_label_new( labname );
	    table_attach( table, widget, 0, row );
	    gtk_misc_set_alignment( GTK_MISC( widget ), 1.0, 0.5 );
	    gtk_widget_show( widget );
	}
	if( entry->type ) {
	    /* 
	       printf( "Prop %s: type %s (0x%x)\n",
	       labname, entry->type->name, (int) entry->type );
	    */
	    if( entry->value )
		entry->widget = VALUE_TO_WIDGET( entry->type, entry->value );
	    else if( entry->deflt )
		entry->widget = VALUE_TO_WIDGET( entry->type, entry->deflt );
	}
	if( entry->widget ) {
	    table_attach( table, entry->widget, 1, row );
	    gtk_widget_show (entry->widget);
	    if( desc )
		gtk_tooltips_set_tip( tooltips, entry->widget, desc, NULL );
	}
	row++;
    }
    return table;
}
