/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: propset.c,v 1.4 1998/05/19 21:08:15 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <proplist.h>
#include <gtk/gtk.h>
#include "wtype.h"
#include "propgroup.h"
#include "propset.h"
#include "proputil.h"

extern int metaPropsChanged;	/* XXX (see malaprop.c) */

PropertySet *PropertySetCreate(
    const char	*name,
    proplist_t	props,
    proplist_t	metaprops
)
{
    PropertySet *set = malloc( sizeof( PropertySet ));

    if( set )
    {
	memset( set, 0, sizeof( PropertySet ));
	set->name = strdup( name );
	set->orig_props = props;
	if( props )
	    PLRetain( props );
	set->orig_metaprops = metaprops;
	if( metaprops )
	    PLRetain( metaprops );
    }
    return set;
}

void PropertySetDestroy( PropertySet *set )
{
    if( set )
    {
	if( set->orig_props )
	    PLRelease( set->orig_props );
	if( set->orig_metaprops )
	    PLRelease( set->orig_metaprops );
	if( set->numGroups )
	{
	    int i;

	    for( i = 0; i < set->numGroups; i++ )
	    {
		PropertyGroupDestroy( set->groups[i] );
	    }
	}
	free( set );
    }
}

static int group_compare( const void *v1, const void *v2 )
{
    PropertyGroup **g1 = (PropertyGroup **) v1;
    PropertyGroup **g2 = (PropertyGroup **) v2;

    return PLOrderCompare( (*g1)->gprops, (*g2)->gprops );
}


void PropertySetSort( PropertySet *set )
{
    if( set )
    {
	int i;

	qsort( set->groups, set->numGroups, sizeof( PropertyGroup * ),
	       group_compare );
	for( i = 0; i < set->numGroups; i++ )
	{
	    PropertyGroupSort( set->groups[i] );
	}
    }
}


PropertyGroup *PropertySetNewGroup( PropertySet *set )
{
    PropertyGroup	*grp;

    if( set->numGroups >= PROP_SET_MAX_GROUPS )
	return NULL;
    if( ( grp = PropertyGroupCreate()) == NULL )
	return NULL;
    set->groups[set->numGroups++] = grp;
    return grp;
}

PropertyGroup *PropertySetFindGroup( PropertySet *set, proplist_t name,
				     proplist_t gprops )
{
    int			i;
    PropertyGroup	*grp;

    /* Look for existing group with this name */
    for( i = 0; i < set->numGroups; i++ )
    {
	grp = set->groups[i];
	if( PLIsEqual( grp->name, name ))
	    return grp;
    }
    /* Not found, create a new one and give it the name */
    grp = PropertySetNewGroup( set );
    if( grp)
    {
	grp->name = PLDeepCopy( name );
	grp->gprops = PLGetDictionaryEntry( gprops, grp->name );
	if( grp->gprops )
	    PLRetain( grp->gprops );
	else
	{
	    grp->gprops = PropertyGroupDefaultProps( grp->name, i + 1 );
	    metaPropsChanged++;
	}
    }
    return grp;
}

GtkWidget *PropertySetWidgetsBuild( PropertySet *set )
{
    GtkWidget	*notebook;
    GtkWidget	*props;
    GtkWidget	*label;
    GtkWidget	*frame;
    GtkWidget	*box;
    int		i;

    notebook = gtk_notebook_new();
    gtk_container_border_width ( GTK_CONTAINER( notebook ), 5);
    gtk_notebook_set_scrollable( GTK_NOTEBOOK( notebook ), TRUE );
    gtk_notebook_set_tab_pos( GTK_NOTEBOOK( notebook ), GTK_POS_RIGHT );
    for( i = 0; i < set->numGroups; i++ )
    {
	/* Build a table of editable widgets */
	props = PropertyGroupWidgetsBuild( set->groups[i] );
	/* Now a frame to surround everything */
	frame = gtk_frame_new( PLGetString( set->groups[i]->name) );
	gtk_widget_show( frame );
	gtk_container_border_width( GTK_CONTAINER( frame ), 10 );
	/* A box to pack things within the frame */
	box = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER( frame ), box );
	gtk_widget_show( box );
	/* Pack the edit widgets at the top */
	gtk_box_pack_start( GTK_BOX( box ), props, FALSE, FALSE, 0 );
	/* And a text box at the bottom */
	label = gtk_label_new(
	    PropertyGroupGetGprop(set->groups[i], "Description" ));
	gtk_box_pack_end( GTK_BOX( box ), label, FALSE, FALSE, 5 );
	gtk_widget_show( label );
	/* Now add the whole thing to the notebook */
	label = gtk_label_new( PLGetString( set->groups[i]->name) );
	gtk_notebook_append_page( GTK_NOTEBOOK(notebook), frame, label );
    }
    gtk_widget_show( notebook );
    return notebook;
}

