/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: wtype.c,v 1.5 1998/06/05 19:21:51 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <proplist.h>
#include <gtk/gtk.h>
#include "wtype.h"
#include "proputil.h"

/*****************************************************************************/
static int GetBoolean( proplist_t pl )
{
    char *s;
    
    if( ! PLIsString( pl ))
	return -1;

    s = PLGetString( pl );

    if( ( s[1]=='\0' && ( s[0] == 'y' ||
			  s[0] == 'Y' ||
			  s[0] == 'T' ||
			  s[0] == 't' )) ||
	( strcasecmp( s, "YES" ) == 0 ||
	  strcasecmp( s, "TRUE" ) == 0 )) {
	return 1;
    }
    if ( ( s[1]=='\0' && ( s[0]=='n' ||
			   s[0]=='N' ||
			   s[0]=='F' ||
			   s[0]=='f' )) ||
	 ( strcasecmp(s, "NO") == 0 ||
	   strcasecmp(s, "FALSE") == 0 )) {
	return 0;
    }
    return -1;
}

static GtkWidget *BoolToWidget( WidgetType *type, proplist_t value )
{
    GtkWidget *w;

    w = gtk_check_button_new ( );
    gtk_toggle_button_set_state( GTK_TOGGLE_BUTTON( w ), GetBoolean( value ));
    return w;
}

static proplist_t BoolFromWidget( WidgetType *type, GtkWidget *widget )
{
    if( widget )
	return (GTK_TOGGLE_BUTTON(widget)->active) ?
	    PLMakeString( "Yes" ) : PLMakeString( "No" );
    return NULL;
}

static int BoolMatch( WidgetType *type, proplist_t value )
{
    if( GetBoolean( value ) >= 0 )
	return TRUE;
    return FALSE;
}
/*****************************************************************************/

static GtkWidget *ColourToWidget( WidgetType *type, proplist_t value )
{
    GtkWidget *w;

    w = gtk_entry_new();
    gtk_entry_set_text( GTK_ENTRY( w ), PLGetDescription( value ));
    gtk_entry_set_position( GTK_ENTRY( w ), 0);
    return w;
}

static proplist_t ColourFromWidget( WidgetType *type, GtkWidget *widget )
{
    if( widget )
	return PLGetProplistWithDescription(
	    gtk_entry_get_text( GTK_ENTRY( widget )) );
    return NULL;
}

static int ColourMatch( WidgetType *type, proplist_t value )
{
    int		i;
    char	*s;
    
    if( PLIsArray( value )) {
	for( i = 0; i < PLGetNumberOfElements( value); i++ ) {
	    if( ! PLIsString( PLGetArrayElement( value, i )))
		return FALSE;
	}
	s = PLGetString( PLGetArrayElement( value, 0 ));
	if( ! strcasecmp( s, "solid" ) ||
	    ! strcasecmp( s, "hgradient" ) ||
	    ! strcasecmp( s, "vgradient" ) ||
	    ! strcasecmp( s, "dgradient" ) ||
	    ! strcasecmp( s, "mhgradient" ) ||
	    ! strcasecmp( s, "mvgradient" ) ||
	    ! strcasecmp( s, "tpixmap" ) ||
	    ! strcasecmp( s, "spixmap" ) ||
	    ! strcasecmp( s, "compose" ))
	    return TRUE;
	return FALSE;
    }
    return FALSE;
}

/*****************************************************************************/

static GtkWidget *ArrayToWidget( WidgetType *type, proplist_t value )
{
    int		i;
    GtkWidget 	*w, *item;

    w = gtk_list_new();
    for( i = 0; i < PLGetNumberOfElements( value ); i++ ) {
	item = gtk_list_item_new_with_label(
	    PLGetString( PLGetArrayElement( value, i )));
	gtk_container_add( GTK_CONTAINER( w ), item );
	gtk_widget_show( item );
    }
    return w;
}

static proplist_t ArrayFromWidget( WidgetType *type, GtkWidget *widget )
{
    proplist_t	array;
    char	*s;
    
    if( widget ) {
	GList *ptr, *list = gtk_container_children( GTK_CONTAINER( widget ));
	GtkWidget *item;
	
	array = PLMakeArrayFromElements( NULL );
	for( ptr = list; ptr; ptr = ptr->next ) {
	    GList *l2;

	    /* Get list item */
	    item = ptr->data;
	    /* Label is first of it's children? */
	    l2 = gtk_container_children( GTK_CONTAINER( item ));
	    gtk_label_get( GTK_LABEL( l2->data ), &s );
	    PLAppendArrayElement( array, PLMakeString( s ));
	    g_list_free( l2 );
	}
	g_list_free( list );
	return array;
    }
    return NULL;
}

static int ArrayMatch( WidgetType *type, proplist_t value )
{
    return PLIsArray( value );
}

/*****************************************************************************/

static GtkWidget *StringListToWidget( WidgetType *type, proplist_t value )
{
    GtkWidget	*w, *menu, *item;
    char	**list = (char **) type->aux;
    char	*valstr = PLGetString( value );
    int		i, index = 0;

    if( ! list )
	return NULL;
    
    w = gtk_option_menu_new();
    menu = gtk_menu_new();
    for( i = 0; *list; list++, i++ )
    {
	if( !strcasecmp( *list, valstr))
	    index = i;		/* Remember index of active option */
	item = gtk_menu_item_new_with_label( *list );
	gtk_object_set_user_data( GTK_OBJECT( item ), (gpointer) *list );
	gtk_menu_append( GTK_MENU( menu ), item );
	gtk_widget_show( item );
    }
    gtk_option_menu_set_menu( GTK_OPTION_MENU( w ), menu);
    gtk_option_menu_set_history( GTK_OPTION_MENU( w ), index);
    return w;
}

static proplist_t StringListFromWidget( WidgetType *type, GtkWidget *widget )
{
    GtkWidget *menu, *label;
    menu = gtk_option_menu_get_menu( GTK_OPTION_MENU( widget ));
    label = gtk_menu_get_active( GTK_MENU( menu ));
    return PLMakeString( (char *)
			 gtk_object_get_user_data( GTK_OBJECT( label )));
    return NULL;
}

static int StringListMatch( WidgetType *type, proplist_t value )
{
    char *s;
    char **list = (char **) type->aux;

    if( ! list )
	return FALSE;
    
    if( PLIsString( value )) {
	s = PLGetString( value );
	for( ; *list; list++ )
	{
	    if( ! strcasecmp( s, *list ))
	    {
		return TRUE;
	    }
	}
    }
    return FALSE;
}

static int StringListDerive( WidgetType *type, proplist_t data )
{
    proplist_t	strings;
    int		i, num;
    char	**auxdata, *str;

    strings = PLGetDict( data, "Strings" );
    if( ! strings )
    {
	fprintf( stderr, "StringList derive %s: No strings list provided\n",
		 type->name );
	return FALSE;
    }
    if( ! PLIsArray( strings ))
    {
	fprintf( stderr, "StringList derive %s: syntax error\n", type->name );
	return FALSE;
    }
    num = PLGetNumberOfElements( strings );
    auxdata = malloc( (num + 1) * sizeof( char *));
    if( ! auxdata )
    {
	fprintf( stderr, "StringList derive %s: malloc error\n", type->name );
	return FALSE;
    }
    for( i = 0; i < num; i++ )
    {
	str = PLGetArrayString( strings, i );
	auxdata[i] = strdup( str );
    }
    auxdata[i] = NULL;
    type->aux = (void *) auxdata;
    return TRUE;
}

/*****************************************************************************/

struct RangeInfo
{
    double	min;
    double	max;
    double	step;
    double	scale;
};

static void NumberChanged( GtkAdjustment *adj, gpointer data )
{
    double newval;
    double slop = adj->step_increment / 100.0;
    /* Find number of step increments from lower bound to value */
    double nstep = ( adj->value - adj->lower ) / adj->step_increment;

    nstep = floor( nstep + 0.5 ); /* Round to nearest integer */
    newval = adj->lower + ( nstep * adj->step_increment );
				/* Convert back to value */
    if( newval > adj->upper ) {
	nstep -= 1;
	newval = adj->lower + ( nstep * adj->step_increment );
    }
    /* printf( "Value: %g %g\n", adj->value, newval ); */
    if( fabs( newval - adj->value ) > slop )
    {
	gtk_adjustment_set_value( adj, newval );
    }
}

static GtkWidget *NumberToWidget( WidgetType *type, proplist_t value )
{
    GtkWidget		*w;
    GtkObject		*adj;
    double		v, min, max, step = 1.0, scale = 1.0, page = 1.0;
    struct RangeInfo	*range = (struct RangeInfo *) type->aux;
    int			digits = 0;

    v = atof( PLGetString( value ));
    if( range )
    {
	scale = range->scale;
	step = range->step;
	min = range->min;
	max = range->max + range->step;
    }
    else
    {
	max = 256.0;
	min = 1.0;
	
	if( v > max )
	    max = (2 * v)+1;
	if( v < min )
	    min = v;
    }
    v /= scale;
    min /= scale;
    max /= scale;
    step /= scale;
    page = (max - min) / 10.0;
    if( step > page )
	page = 2 * step;
    adj = gtk_adjustment_new( v, min, max, step, page, step );
    gtk_signal_connect( GTK_OBJECT( adj ), "value_changed",
			(GtkSignalFunc) NumberChanged, NULL );
    /* Set the value again, so NumberChanged gets a look in */
    gtk_adjustment_set_value( GTK_ADJUSTMENT( adj ), v );
    w = gtk_hscale_new( GTK_ADJUSTMENT( adj ));
    gtk_widget_set_usize( GTK_WIDGET( w ), 150, 30 );
    gtk_range_set_update_policy( GTK_RANGE( w ), GTK_UPDATE_CONTINUOUS );
    /* Attempt to use a useful number of significant digits */
    if( fmod( step, 1.0 ) != 0.0 )
	digits++;
    if( fmod( step * 10.0, 1.0 ) != 0.0 )
	digits++;
    if( fmod( step * 100.0, 1.0 ) != 0.0 )
	digits++;
    gtk_scale_set_digits( GTK_SCALE( w ), digits );
    gtk_scale_set_draw_value( GTK_SCALE( w ), TRUE );

    return w;
}

static proplist_t NumberFromWidget( WidgetType *type, GtkWidget *widget )
{
    struct RangeInfo	*range = (struct RangeInfo *) type->aux;
    double scale = range ? range->scale : 1.0;

    if( widget )
    {
	GtkAdjustment *adj = GTK_RANGE(widget)->adjustment;
	char buffer[64];

	sprintf( buffer, "%g", adj->value * scale );
	return PLMakeString( buffer );
    }
    return NULL;
}

static int NumberMatch( WidgetType *type, proplist_t value )
{
    if( PLIsString( value ))
    {
	char buffer[64];
	double v = atof( PLGetString( value ));

	sprintf( buffer, "%g", v );
	/* printf( "Compare: '%s', '%s'\n", buffer, PLGetString( value )); */
	return( ! strcmp( buffer, PLGetString( value )));
    }
    return FALSE;
}

static int NumberDerive( WidgetType *type, proplist_t data )
{
    struct RangeInfo	*range = malloc( sizeof( struct RangeInfo ));

    range->min = PLGetDictDouble( data, "Min", 1.0 );
    range->max = PLGetDictDouble( data, "Max", 1.0 );
    range->step = PLGetDictDouble( data, "Step", 1.0 );
    range->scale = PLGetDictDouble( data, "Scale", 1.0 );
    
    type->aux = (void *) range;
    return TRUE;
}

/*****************************************************************************/

static GtkWidget *StringToWidget( WidgetType *type, proplist_t value )
{
    GtkWidget *w;

    w = gtk_entry_new();
    gtk_entry_set_text( GTK_ENTRY( w ), PLGetString( value ));
    gtk_entry_set_position( GTK_ENTRY( w ), 0);
    return w;
}

static proplist_t StringFromWidget( WidgetType *type, GtkWidget *widget )
{
    if( widget )
	return PLMakeString( gtk_entry_get_text( GTK_ENTRY( widget )) );
    return NULL;
}

static int StringMatch( WidgetType *type, proplist_t value )
{
    return PLIsString( value );
}

/*****************************************************************************/


static WidgetType base_wtypes[] = {
    { "Boolean", 10, BoolToWidget, BoolFromWidget, BoolMatch, NULL, NULL },
    { "WMColour", 20, ColourToWidget, ColourFromWidget, ColourMatch,
      NULL, NULL },
    { "Array", 30, ArrayToWidget, ArrayFromWidget, ArrayMatch, NULL, NULL },
    { "StringList", 40, StringListToWidget, StringListFromWidget,
      StringListMatch, StringListDerive, NULL },
    { "Number", 50, NumberToWidget, NumberFromWidget, NumberMatch,
      NumberDerive, NULL },
    { "String", 100, StringToWidget, StringFromWidget, StringMatch, NULL,NULL},
    { NULL },
};

#define WINIT()     if( !winitted ) winit();

static int winitted = 0;
static int nTypeAlloc = 0;
static int nType = 0;
static WidgetType *wtypes;

static WidgetType *walloc()
{
    WidgetType	*array;
    int		size;
    
    if( nType >= nTypeAlloc )
    {
	size = nTypeAlloc + 16;
	if( wtypes )
	    array = realloc( wtypes, size * sizeof( WidgetType ));
	else
	    array = malloc( size * sizeof( WidgetType ));
	if( array )
	{
	    wtypes = array;
	    nTypeAlloc = size;
	}
	else
	{
	    return NULL;
	}
    }
    return( &wtypes[nType++] );
}

static int wregister( WidgetType *wt )
{
    WidgetType *wnew;
    
    if(  ( !wt ) || ( !wt->name ))
	return FALSE;

    wnew = walloc();
    if( ! wnew )
	return FALSE;

    memcpy( wnew, wt, sizeof( WidgetType ));
    wnew->pname = PLMakeString( wnew->name );
    
    return TRUE;
}

static int wcompare( const void *v1, const void *v2 )
{
    WidgetType *w1 = (WidgetType *) v1;
    WidgetType *w2 = (WidgetType *) v2;

    if( w1->priority > w2->priority )
	return 1;
    if( w1->priority < w2->priority )
	return -1;
    return 0;
}

static void wsort()
{
    qsort( wtypes, nType, sizeof( WidgetType ), wcompare );
}

static void winit()
{
    WidgetType *t;

    for( t = base_wtypes; t->name; t++ ) {
	wregister( t );
    }
    wsort();
    winitted = 1;
}

WidgetType *WidgetTypeByName( const char *name )
{
    WidgetType	*t;
    int		i;

    WINIT();

    for( i = 0, t = wtypes; i < nType; i++, t++ )
    {
	if( !strcmp( t->name, name ))
	    return t;
    }
    return NULL;
}


WidgetType *WidgetTypeInfer( proplist_t value )
{
    WidgetType	*t;
    int		i;

    WINIT();

    for( i = 0, t = wtypes; i < nType; i++, t++ )
    {
	if( TYPE_MATCH( t, value ))
	    return t;
    }
    return NULL;
}

int WidgetTypeDerive( char *basename, char *typename, proplist_t data )
{
    WidgetType *base, w;
    WINIT();

    /* printf( "Derive %s from %s\n", typename, basename ); */
    if( WidgetTypeByName( typename ))
    {
	fprintf( stderr, "Cannot override definition of existing type %s\n",
		 typename );
	return FALSE;
    }
    base = WidgetTypeByName( basename );
    if( ! base )
    {
	fprintf( stderr, "No such base class %s\n", basename );
	return FALSE;
    }
    if( base->derive )
    {
	/* Copy base class and change name */
	memcpy( &w, base, sizeof( WidgetType ));
	w.name = strdup( typename ); /* XXX handle NULL return */
	if( (base->derive)( &w, data ) )
	{
	    wregister( &w );
	    wsort();
	}
	else
	{
	    fprintf( stderr, "Derive %s from %s failed\n", typename, basename);
	    return FALSE;
	}
    }
    else
    {
	fprintf( stderr, "Base class %s may not be derived\n", basename );
	return FALSE;
    }
    return TRUE;
}
